/**
 * @author Olgierd Humeńczuk olgierdh@gmail.com
 * @file main
 */

#include <iostream>
#include <stdlib.h>
#include <stdexcept>
#include <vector>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <SDL.h>
#else
#include <SDL/SDL.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#undef main

/*static const GLfloat cubeArray[][3] =
{
    {-0.5f, -0.5f, -0.5f}, {-0.5f, 0.5f, -0.5f}, {0.5f, 0.5f, -0.5f}, {0.5f, -0.5f, -0.5f},
    {-0.5f, -0.5f, 0.5f}, {0.5f, -0.5f, 0.5f}, {0.5f, 0.5f, 0.5f}, {-0.5f, 0.5f, 0.5f},
    {-0.5f, -0.5f, -0.5f}, {0.5f, -0.5f, -0.5f}, {0.5f, -0.5f, 0.5f}, {-0.5f, -0.5f, 0.5f},
    {-0.5f, 0.5f, -0.5f}, {-0.5f, 0.5f, 0.5f}, {0.5f, 0.5f, 0.5f}, {0.5f, 0.5f, -0.5f},
    {-0.5f, 0.5f, -0.5f}, {-0.5f, -0.5f, -0.5f}, {-0.5f, -0.5f, 0.5f}, {-0.5f, 0.5f, 0.5f},
    {0.5f, -0.5f, -0.5f}, {0.5f, 0.5f, -0.5f}, {0.5f, 0.5f, 0.5f}, {0.5f, -0.5f, 0.5f}
};

//deklaracja wierzchołków płaszczyzny
static const GLfloat planeArray[][3] =
{
    { -0.5f, -0.5f, -0.5f }, { -0.5f, -0.5f, 0.5f }, { 0.5f, -0.5f, 0.5f }, { 0.5f, -0.5f, -0.5f }
};*/



/**
 * @struct position2d
 * @brief pozycja dwuwymiarowa
 */
struct position2d
{
    /**
     * @brief domyslny konstr.
     * inicjalizuje wsp. wartosciami zerowymi
     */
    position2d( void ) : x( 0 ), y( 0 ) { }

    /// wspolrzedna x
    int x;

    /// wspolrzedna y
    int y;

} mousePosition;

/**
 * @brief Funkcja przetwarza zdarzenia pochodzace z kontrolerów takich jak:
 *          * klawiatura
 *          * mysz komputerowa
 *
 * @return true jeżeli aplikacja ma pozostać otwarta, false wpp
 */
bool processInput( void )
{
    // zmienna pomocnicza
    SDL_Event event;

    // czy są jakieś czekające na przetworzenie zdarzenia ?
    if ( SDL_PollEvent( &event ) )
    {
        // sprawdzenie typu zdarzenia
        switch( event.type )
        {
            // wyjście z aplikacji
            case SDL_QUIT:
                return false;

            // użytkownik nacisnął któryś z klawiszy na klawiaturze
            case SDL_KEYDOWN:
                // jeżeli nacisnał klawisz eskape zamykamy aplikacje
                if( event.key.keysym.sym == SDLK_ESCAPE )
                {
                    return false;
                }
                break;
            case SDL_MOUSEMOTION:
                SDL_GetMouseState( &mousePosition.x, &mousePosition.y );
                break;
        }
    }
    return true;
}
/**
 * @brief Funkcja odpowiadająca za inicjalizację biblioteki SDL
 *          oraz utworzenie okna z kontekstem OpenGL.
 * @param scrWidth szerokosc ekranu
 * @param scrHeight wysokosc ekranu
 *
 * @throw std::runtime_error jeżeli nastąpi jakikolwiek błąd.
 */
void initSDL( int scrWidth, int scrHeight )
{
    // zaincjalizowanie części biblioteki odpowiadającej za wyświetlanie
    if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
    {
        throw std::runtime_error( "Can't initialize SDL\n" );
    }

    // ustawienie ilości bitów dla każdego z kolorów
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );

    // ustawienie ilości bitów dla specjalnych buforów
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
    SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );

    // włączenie podwójnego buforowania
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, true );
    SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );

    // stworzenie okna aplikacji z kontekstem OpenGL'a
    if ( SDL_SetVideoMode( scrWidth, scrHeight, 24, SDL_OPENGL ) == 0 )
    {
        throw std::runtime_error(
            std::string( "Unable to set video mode: " ).append(
            	std::string( SDL_GetError() ) ) );
    }
}

/**
 * @brief ustawia zrodlo swiatla punktowego
 * @param id numer swiatla
 * @param position pozycja światła w przestrzeni
 * @param ambient_col kolor światła otoczenia
 * @param diffuse_col kolor światła rozproszonego
 * @param specular_col kolor światła odbitego
 */
void setPointLight(
    GLenum id,
    GLfloat position[ 4 ],
    GLfloat ambient_col[ 4 ],
    GLfloat diffuse_col[ 4 ],
    GLfloat specular_col[ 4 ] )
{
    glEnable( id );

    glLightfv( id, GL_AMBIENT, ambient_col );
    glLightfv( id, GL_DIFFUSE, diffuse_col );
    glLightfv( id, GL_SPECULAR, specular_col );
    glLightfv( id, GL_POSITION, position );
}

/**
 * @brief funkcja ustawia parametry materiału
 * @param ambient_col kolor materiału nieoświetlonego żadnym światłem
 * @param diffuse_col w jaki sposób materiał odbija światło rozproszone
 * @param specular_col w jaki sposób materiał odbija światło odbite
 * @param shiness połyskliwość materiału
 */
void setMaterial(
    GLfloat ambient_col[ 4 ],
    GLfloat diffuse_col[ 4 ],
    GLfloat specular_col[ 4 ],
    GLfloat shiness )
{
    glMaterialfv( GL_FRONT, GL_AMBIENT, ambient_col );
    glMaterialfv( GL_FRONT, GL_DIFFUSE, diffuse_col );
    glMaterialfv( GL_FRONT, GL_SPECULAR, specular_col );

    glMaterialf( GL_FRONT, GL_SHININESS, shiness );
}

/**
 * @brief Iniclalizacja stanu poczatkowego niektorych
 *         wartosci OpenGL.
 * @param width szerokosc ekranu
 * @param height wysokosc ekranu
 */
void initOpenGL( int width, int height )
{
    glMatrixMode( GL_PROJECTION );
    gluPerspective( 45.0f, width / ( GLfloat ) height, 0.1f, 100.0f );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    //włączenie testowania bufora głębokości
    glEnable( GL_DEPTH_TEST );

    //sposób rasteryzacji wielokątów
    glPolygonMode( GL_FRONT, GL_FILL );

    //włączenie liczenia oświetlenia
    glEnable( GL_LIGHTING );

}

/**
 * @struct vector3f
 * @brief Struktura przechowująca współrzędne wektora
 */
struct vector3f
{

    /**
     * @brief Domyslny konstruktor
     * Ustawia wszystkie wspolrzedne na ( 0, 0, 0 )
     */
    vector3f( void ) : x( .0f ), y( .0f ), z( .0f )
    {
    }

    /**
     * @brief Konstruktor ustawiajacy wspolrzedne
     * @param in_x wspolrzedna x
     * @param in_y wspolrzedna y
     * @param in_z wspolrzedna z
     */
    vector3f( GLfloat in_x, GLfloat in_y, GLfloat in_z )
        : x( in_x ), y( in_y ), z( in_z )
    {
    }

    ///wspolrzedna x
    GLfloat x;
    ///wspolrzedna y
    GLfloat y;
    ///wspolrzedna z
    GLfloat z;
};

/**
 * @struct vertex3f
 * @brief Struktura opisujaca wierzchołek w przestrzeni 3d
 */
struct vertex3f
{
    typedef vector3f position_type;
    typedef vector3f normal_type;

    /**
     * @brief Domyslny konstruktor
     * Ustawia wszystkie wektory na ( 0, 0, 0 )
     */
    vertex3f( void ) { }

    /**
     * @brief Konstruktor ustawiajacy wspolrzedne
     * @param pos pozycja wierzchołka w przestrzeni
     * @param n normalna wierzcholka
     */
    explicit vertex3f( const position_type& pos, const normal_type& n  )
        : position( pos ), normal( n )
    {
    }

    ///pozycja wierzchołka
    position_type position;

    ///normalna w wierzchołku
    normal_type normal;
};

/** Deklaracja typu tablicy wierzcholkow */
typedef std::vector< vertex3f > vertexes;

/**
 * @brief Tworzy tablice wierzcholkow szescianu
 * @param width dlugosc boku szescianu
 */
vertexes createCube( GLfloat width )
{
    vertexes ret;
    //rezerwacja pamięci
    ret.reserve( 6 * 4 * sizeof( vertexes::value_type ) );

    //zapamiętanie wartości pomocniczych
    GLfloat min = -width * 0.5f;
    GLfloat max = -min;

    //przednia sciana
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( .0f, .0f, 1.0f ) ) );

    //tylnia sciana
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( .0f, .0f, -1.0f ) ) );

    //prawa sciana
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( 1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( 1.0f, .0f, .0f ) ) );

    //lewa sciana
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( -1.0f, .0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( -1.0f, .0f, .0f ) ) );

    //gorna sciana
    ret.push_back(
        vertex3f( vector3f( min, max, min ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, max, max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, max ),
                  vector3f( .0f, 1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, max, min ),
                  vector3f( .0f, 1.0f, .0f ) ) );

    //dolna sciana
    ret.push_back(
        vertex3f( vector3f( min, min, max ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( min, min, min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, min ),
                  vector3f( .0f, -1.0f, .0f ) ) );
    ret.push_back(
        vertex3f( vector3f( max, min, max ),
                  vector3f( .0f, -1.0f, .0f ) ) );

    return ret;
}

/**
 * @brief Funkcja pomocnicza, przelicza pozycję z jednej przestrzeni na inna
 *        okreslona podanymi parametrami
 *
 * @param posX pozycja wejściowa
 * @param maxPosX maksymalna wartosc jaką może uzyskać parametr posX
 * @param dstPosMin minimalna wartosc pozycji w docelowej przestrzeni
 * @param dstPosMax maksymalna wartosc pozycji w docelowej przestrzeni
 */
GLfloat transformPos( GLfloat posX, GLfloat maxPosX,
							 GLfloat dstPosMin, GLfloat dstPosMax )
{
    GLfloat normalized = posX / maxPosX;
    return dstPosMin + ( ( dstPosMax - dstPosMin ) * normalized );
}

/**
 * @brief Główna metoda programu ( entry point )
 * @param argc ilość argumentów przekazanych do programu
 * @param argv tablica wartości argumentów
 */
int main( const int /*argc*/, const char** /*argv*/ )
{
	 const int screenX = 800;
	 const int screenY = 600;

    try
    {
        initSDL( screenX, screenY );
        initOpenGL( screenX, screenY );

        vertexes cubeVertexes = createCube( 2.0f );

        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );

        float angle = .0f;

        // główna pętla
        while( processInput() )
        {
            glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
            glMatrixMode( GL_MODELVIEW );
            glLoadIdentity();

            glPushMatrix();
            gluLookAt( 0.0f, 0.0f, 20.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );

            {
                //dane światła
                GLfloat position[] = { 0.0f, 10.0f, 10.0f, 1.0f };
                GLfloat ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
                GLfloat diffuse[] = { 0.8f, 0.8f, 0.8f, 1.0f };
                GLfloat specular[] = { 0.8f, 0.8f, 0.8f, 1.0f };

                setPointLight( GL_LIGHT0, position, ambient, diffuse, specular );
            }

            glPushMatrix();
            glTranslatef(
                transformPos( static_cast< GLfloat >( mousePosition.x ),
			      static_cast< GLfloat >( screenX ), -5.5f, 5.5f ),
                0.0f, 0.0f );

            glRotatef( angle, 1.0f, .0f, .0f );

            //prędkość obrotu
            angle += 0.4f;

            glVertexPointer( 3, GL_FLOAT, sizeof( vertexes::value_type ), &cubeVertexes[ 0 ] );
            glNormalPointer( GL_FLOAT, sizeof( vertexes::value_type ), &cubeVertexes[ 0 ].normal );

            {
                GLfloat ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
                GLfloat diffuse[] = { 1.0f, 0.0f, 0.0f, 1.0f };
                GLfloat specular[] = { 0.2f, 0.2f, 0.2f, 1.0f };

                setMaterial( ambient, diffuse, specular, 0.9f );
            }

            glDrawArrays( GL_QUADS, 0, cubeVertexes.size() );

            glPopMatrix();
            glPopMatrix();

            SDL_GL_SwapBuffers();
        }
    }
    catch( const std::exception& e )
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
    }

    SDL_Quit();
    return 0;
}
